require 'net/ssh'
require 'net/scp'

require "remote_marshal/version"

class RemoteMarshal
  def initialize(host, user, options)
    @ssh = create_ssh(host, user, options)
  end

  def eval(&block)
    command = block.to_source(strip_enclosure: true)

    call { |dir| marshal(StringIO.new(command), File.join(dir, token)) }
  end

  def file(filename)
    call { |dir| marshal(filename, File.join(dir, File.basename(filename))) }
  end

  def exec!(command)
    status, output, error = nil, [], []

    @ssh.call do |ssh|
      channel = ssh.open_channel do |chan|
        chan.exec command do |ch, success|
          raise command unless success

          ch.on_data                   { |_, data|    output << data }
          ch.on_extended_data          { |_, _, data| error  << data }
          ch.on_request('exit-status') { |_, data|    status =  data.read_long }
        end
      end

      channel.wait
    end

    [status] + [output, error].map(&:join)
  end

  %i(upload! download!).each do |name|
    define_method(name) do |src, dst|
      @ssh.call { |ssh| ssh.scp.public_send(name, src, dst) }
    end
  end

  def test?(name, option = nil)
    exec!(%([ #{option} #{name} ]))
  end

  private

  def call(&block)
    status, output, error = mktmpdir

    return [status, output, error] unless status == 0

    dir                   = output.chomp
    ret                   = block.(dir)
    status, output, error = remove_entry_secure(dir)

    return [status, output, error] unless status == 0

    status, output, error = *ret

    return [status, output, error] unless status == 0

    [status, Marshal.load(output.chomp), error]
  end

  def mktmpdir
    exec!("ruby -r tmpdir -e 'puts Dir.mktmpdir'")
  end

  def remove_entry_secure(dir)
    exec!(%(ruby -r fileutils -e "FileUtils.remove_entry_secure '#{dir}'"))
  end

  def token
    seed = ([Time.now] + (1..10).map { rand.to_s }).join('--')

    Digest::SHA1.hexdigest(seed)
  end

  def marshal(src, dst)
    upload!(src, dst)

    exec!("ruby -e 'puts Marshal.dump(eval(open(ARGV[0], &:read)))' #{dst}")
  end

  def create_ssh(host, user, options)
    lambda do |&block|
      Net::SSH.start(host, user, options) do |ssh|
        block.(ssh)

        ssh.loop
      end
    end
  end
end
